﻿/**

Livewire Flash - Core code for running the Livewire algorithm
Copyright (C) 2011  Jeffrey Bush  jeff@coderforlife.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

**/

package Livewire
{	
	import Livewire.Pools.Points;
	import Livewire.Pools.NumberVectors;

	import flash.geom.Point;
	
	public final class LivewireCalculator extends Threaded
	{
		private static const SQRT2:Number = 1.41421356237309504880168872420969807856967187537694807317667973799;
		private function I(x:uint, y:uint):uint { return x + y * this._w; }

		/// <summary>The status of the livewire algorithm for a pixel</summary>
		//final class Status 
		//{
			/// <summary>The livewire algorithm has not reached the pixel</summary>
			private static const StatusUnvisited:uint = 0; 
			/// <summary>The livewire algorithm has finished calculating the pixel</summary>
			private static const StatusInEdge:uint = 1; 
			/// <summary>The livewire algorithm has reached the pixel but not finished calculating it</summary>
			private static const StatusVisited:uint = 2; 
		///}

		/// <summary>The width of the data</summary>
		private var _w:uint;

		/// <summary>The height of the data</summary>
		private var _h:uint;

		/// <summary>The scale of the data (amount of binning)</summary>
		private var _scale:uint;

		/// <summary>The weight calculator backing this livewire calculator</summary>
		private var _weights_calc:WeightCalculator;

		/// <summary>The weights used in the livewire algorithm</summary>
		private var _weights:Vector.<uint>;
		
		/// <summary>The X coordinate of the point at which the livewire data is calculated for</summary>
		private var _x:uint;
		
		/// <summary>The Y coordinate of the point at which the livewire data is calculated for</summary>
		private var _y:uint;

		/// <summary>The point at which the livewire data is calculated for (in the form x+y*w)</summary>
		private var _I:uint;

		/// <summary>The livewire algorithm status for every pixel</summary>
		private var _status:Vector.<uint>;

		/// <summary>The livewire algorithm trace: the point which led to the pixel's best score (in the form x+y*w)</summary>
		private var _trace:Vector.<uint>;

		/// <summary>The priority queue representing the "InEdge" points used for calculating the data, which we only want to allocated once</summary>
		private var _edge:PointPriorityQueue;
		
		// Object Pools - Unique to Flash
		private var _points:Points;
		private var _vectors:NumberVectors;

		/// <summary>Create a livewire calculator object from the given weights.</summary>
		/// <param name="weights">The weights object, which includes the width and height of the output data.</param>
		/// <remarks>The weight matrix doesn't have to be done calculating when given to the livewire calculator.</remarks>
		public final function LivewireCalculator(weights:WeightCalculator)
		{
			super("Livewire Calculator");
			this._weights_calc = weights;
			this._weights = weights.GetWeights();
			this._w = weights.GetReducedWidth();
			this._h = weights.GetReducedHeight();
			this._scale = weights.GetScale();
			this._trace = new Vector.<uint>(this._w*this._h, true);
			this._edge = new PointPriorityQueue(this._w, this._h);
			this._points = new Points(this._w+this._w+this._h+this._h);
			this._vectors = new NumberVectors(4);
			this.SetTotalProgress(this._w*this._h);
		}
		
		/// <summary>Starts the livewire-calculating thread for the given point, see prepareLivewire for the function that actually does the work</summary>
		/// <param name="x">The X coordinate at which to calculate the livewire data for, it is given a score of 0</param>
		/// <param name="y">The Y coordinate at which to calculate the livewire data for, it is given a score of 0</param>
		public final function Start(x:uint, y:uint):void
		{
			this._x = x / this._scale;
			this._y = y / this._scale;
			this._I = I(this._x, this._y);
			this.start();
		}
		
		protected final override function StepFirst(context:*):void
		{
			// Reset data structures
			this._status = new Vector.<uint>(this._w * this._h, true);
			this._edge.Clear();

			// Start with just the start point
			this._edge.Enqueue(this._x, this._y, this._I, 0);
			//this._status[this._I] = StatusInEdge; // not necessary, immediately set to visited
			this._trace[this._I] = int.MAX_VALUE;
		}

		/// <summary>Run the livewire LivewireCalculator</summary>
		protected final override function Step(iters:uint, context:*):Boolean
		{
			for (var iter:uint = 0; iter < iters; ++iter)
			{
				// Get the best-scoring point and mark it as visited (done being calculated)
				var e:* = this._edge.Dequeue();
				if (e == null) return false;
				var X:uint = e.x, Y:uint = e.y, I:uint = e.I, S:uint = e.score;
				this._edge.ReturnEntry(e);
				
				this._status[I] = StatusVisited;
				
				const w:uint = this._w, w1:uint = w-1, h:uint = this._h, h1:uint = h-1;
				
	//#define CALC_POINT(dx, dy, diag)	this.CalcPoint(X + dx, Y + dy, I + dy*w + dx, diag, I, S)
				if (X && X < w1 && Y && Y < h1)
				{
					// entire 3x3 neighborhood
					this.CalcPoint(X - 1, Y - 1, I - w - 1, true,  I, S);
					this.CalcPoint(X,     Y - 1, I - w,     false, I, S);
					this.CalcPoint(X + 1, Y - 1, I - w + 1, true,  I, S);
					this.CalcPoint(X - 1, Y,     I     - 1, false, I, S);
					this.CalcPoint(X + 1, Y,     I     + 1, false, I, S);
					this.CalcPoint(X - 1, Y + 1, I + w - 1, true,  I, S);
					this.CalcPoint(X,     Y + 1, I + w,     false, I, S);
					this.CalcPoint(X + 1, Y + 1, I + w + 1, true,  I, S);
				}
				else
				{
					// Get the range for the neighboring points
					const min_x:uint = X ? X - 1 : 0, max_x:uint = X == w1 ? w1 : X + 1;
					const min_y:uint = Y ? Y - 1 : 0, max_y:uint = Y == h1 ? h1 : Y + 1;
		
					// Cycle through all neighboring points
					for (var y:uint = min_y, off = y * w; y <= max_y; ++y, off += w)
						for (var x:uint = min_x, i = off + x; x <= max_x; ++x, ++i)
							this.CalcPoint(x, y, i, ((X-x + Y-y) & 1) == 0, I, S);
				}
			}
			return true;
		}

		private final function CalcPoint(x:uint, y:uint, i:uint, diagonal:Boolean, I:uint, S:uint):void
		{
			// Make sure the point isn't already done
			const status:uint = this._status[i];
			if (status != StatusVisited)
			{
				// Calculate the score for this point
				const s:uint = S + (diagonal ? (int)(SQRT2 * this._weights[i]) : this._weights[i]);
		
				// Add the point to the edge or update its score
				if (status == StatusUnvisited)
				{
					this._edge.Enqueue(x, y, i, s);
					this._status[i] = StatusInEdge;
					this._trace[i] = I;
				}
				else if (this._edge.DescreaseScore(i, s))
				{
					this._trace[i] = I;
				}
			}
		}
		
		public final function GetTrace(x:uint, y:uint):Vector.<Point>
		{
			const scale:uint = this._scale, s2:uint = scale / 2;
			var pts:Vector.<Point> = new Vector.<Point>();
			var I:uint = I(x / scale, y / scale);
			if (this._status != null && this._status[I] == StatusVisited) // The endpoint has been solved, so we can actually draw the livewire
			{
				// Get every point from end to start by looping through the trace data
				do
				{
					var Y:int = I / this._w;
					pts.push(this._points.Get((I - this._w * Y) * scale + s2, Y * scale + s2));
				}
				while ((I = this._trace[I]) != int.MAX_VALUE); // (I = this._trace[I]) != this._I
			}
			return pts;
		}
		
		public final function ReturnTrace(pts:Vector.<Point>):void
		{
			for (var i:uint; i < pts.length; ++i)
				this._points.Return(pts[i]);
		}
		
		public final function GetRawTrace(x:uint, y:uint):Vector.<Number>
		{
			const scale:uint = this._scale, s2:uint = scale / 2;
			var pts:Vector.<Number> = this._vectors.Get(0, false);
			var I:uint = I(x / scale, y / scale);
			if (this._status != null && this._status[I] == StatusVisited) // The endpoint has been solved, so we can actually draw the livewire
			{
				// Get every point from end to start by looping through the trace data
				do
				{
					var Y:int = I / this._w;
					pts.push((I - this._w * Y) * scale + s2);
					pts.push(Y * scale + s2);
				}
				while ((I = this._trace[I]) != int.MAX_VALUE); // (I = this._trace[I]) != this._I
			}
			return pts;
		}
		
		public final function ReturnRawTrace(pts:Vector.<Number>):void
		{
			this._vectors.Return(pts);
		}
	}
}
