﻿/**

Livewire Flash - Core code for running the Livewire algorithm
Copyright (C) 2011  Jeffrey Bush  jeff@coderforlife.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

**/


/*
 * Inspired by a min-heap priority queue by Alexey Kurakin (http://www.codeproject.com/KB/threads/PriorityQueueGeneric.aspx)
 */

package Livewire
{
	internal final class PointPriorityQueue
	{
		/* Necessary public functions:
		 *  Enqueue(x:uint, y:uint, I:uint, score:uint):void
		 *  Dequeue():Entry
		 *  DescreaseScore(I:uint, score:uint):Boolean
		 *  Clear():void
		 *  ReturnEntry(e:Entry):void       (Unique to Flash, see EntryObjectPool)
		 * All other public functions can be removed
		 */
		
		private var _width:uint, _height:uint;
		private var _heap:Vector.<Entry>, _map:Vector.<Entry>;
		private var _count:uint = 0;
		
		private var _pool:EntryPool;
		
		public final function PointPriorityQueue(w:uint, h:uint)
		{
			this._width = w;
			this._height = h;
			
			// a great guess for maxium edge size is the perimeter
			// it is possible to get much larger though that is rare
			var perim:uint = w+h+w+h;
			
			this._heap = new Vector.<Entry>(perim, true);
			this._map = new Vector.<Entry>(w*h, true);
			this._pool = new EntryPool(perim);
		}
		
//		//public final function Enqueue(I:uint, score:uint):void { this.Insert(I % this._width, I / this._width, I, score); }
//		//public final function Enqueue(x:uint, y:uint, score:uint):void { this.Insert(x, y, x + y * this._width, score); }
		public final function Enqueue(x:uint, y:uint, I:uint, score:uint):void { this.Insert(x, y, I, score); }

		public final function Dequeue():Entry
		{
			if (this._count == 0) return null;
			const e:Entry = this._heap[0];
			this.DeleteRoot();
			return e;
		}
		
		public final function ReturnEntry(e:Entry):void { this._pool.Return(e); }
		
//		public final function Peek():Entry { return this._count == 0 ? null : this._heap[0]; }
		
//		public final function IsEmpty():Boolean { return this._count == 0; }
//		public final function Size():uint { return this._count; }
		
		public final function Clear():void
		{
			for (var i:uint = 0; i < this._count; ++i)
			{
				var e:Entry = this._heap[i];
				this._map[e.I] = null;
				this._heap[i] = null;
				this._pool.Return(e);
			}
			this._count = 0;
			//CHECK(true);
		}
		
//		//public final function Contains(x:uint, y:uint):Boolean { return this._map[x + y * this._width] != null; }
//		//public final function UpdateScore(x:uint, y:uint, score:uint):Boolean { return this.UpdateScore(x + y * this._width, score); }
//		//public final function DescreaseScore(x:uint, y:uint, score:uint):Boolean { return this.DescreaseScore(x + y * this._width, score); }
//		//public final function IncreaseScore(x:uint, y:uint, score:uint):Boolean { return this.IncreaseScore(x + y * this._width, score); }
		
//		public final function Contains(I:uint):Boolean { return this._map[I] != null; }
//		public final function UpdateScore(I:uint, score:uint):Boolean
//		{
//			//ASSERT(this._map[I] != null);
//			const i:uint = this._map[I].index, s:uint = this._heap[i].score;
//			if (score == s) return false;
//			this._heap[i].score = score;
//			if (score > s)	this.HeapifyFromBeginningToEnd(i); // priority is larger, heapify
//			else			this.HeapifyFromEndToBeginning(i); // priority is smaller, heapify
//			return true;
//		}
		public final function DescreaseScore(I:uint, score:uint):Boolean
		{
			//ASSERT(this._map[I] != null);
			const i:uint = this._map[I].index;
			if (score >= this._heap[i].score) return false;
			this._heap[i].score = score;
			this.HeapifyFromEndToBeginning(i);
			return true;
		}
//		public final function IncreaseScore(I:uint, score:uint):Boolean
//		{
//			//ASSERT(this._map[I] != null);
//			const i:uint = this._map[I].index;
//			if (score <= this._heap[i].score) return false;
//			this._heap[i].score = score;
//			this.HeapifyFromBeginningToEnd(i);
//			return true;
//		}
		
		private final function ExchangeElements(a:uint, b:uint):void
		{
			const e:Entry = this._heap[a];
			(this._heap[a] = this._heap[b]).index = a;
			(this._heap[b] = e).index = b;
			//CHECK(false);
		}
		
		private final function Insert(x:uint, y:uint, I:uint, score:uint):void
		{
			//ASSERT(x < this._width && y < this._height);
			//ASSERT(this._map[I] == null);
		
			const count:uint = this._count;
			if (count == this._heap.length)
			{
				this._heap.fixed = false;
				this._heap.length = count * 2;
				this._heap.fixed = true;
			}
			this._heap[count] = this._map[I] = this._pool.Get(score, count, x, y, I);
			++this._count;
			//CHECK(false);
		
			// heapify after insert, from end to beginning
			this.HeapifyFromEndToBeginning(count);
		}
		private final function HeapifyFromEndToBeginning(pos:uint):void
		{
			//ASSERT(pos < this._count); 
			while (pos > 0) {
				// heap[i] have children heap[2*i + 1] and heap[2*i + 2] and parent heap[(i-1)/ 2];
				var parentPos:uint = (pos - 1) / 2;
				if (this._heap[parentPos].score > this._heap[pos].score)
				{
					this.ExchangeElements(parentPos, pos);
					pos = parentPos;
				}
				else break;
			}
			//CHECK(true);
		}
		
		private final function DeleteRoot():void
		{
			//ASSERT(count > 0); 
		
			this._map[this._heap[0].I] = null;
			
			if (--this._count > 0)
			{
				(this._heap[0] = this._heap[this._count]).index = 0;

				//CHECK(false);
				
				// heapify
				this.HeapifyFromBeginningToEnd(0);
			}
			
			this._heap[this._count] = null;
		}
		private final function HeapifyFromBeginningToEnd(pos:uint):void
		{
			const count:uint = this._count;
		
			//ASSERT(pos < count); 
			
			const S:uint = this._heap[pos].score;
			var smallest:uint = pos;
			while (true)
			{
				// on each iteration exchange element with its smallest child
				// heap[i] have children heap[2*i + 1] and heap[2*i + 2] and parent heap[(i-1)/ 2];
				var score:uint, l:uint = pos * 2 + 1, r:uint = pos * 2 + 2;
				if (l < count && S > this._heap[l].score) { smallest = l; score = this._heap[l].score; } else { score = S; }
				if (r < count && score > this._heap[r].score) { smallest = r; }
				else if (smallest == pos) break;

				this.ExchangeElements(smallest, pos);
				pos = smallest;
			}
		
			//CHECK(true);
		}
	}
}

internal final class Entry
{
	public var score:uint;
	public var index:uint;
	public var x:uint, y:uint, I:uint;
	public final function Entry(s:uint, index:uint, x:uint, y:uint, I:uint)
	{
		this.score = s;
		this.index = index;
		this.x = x;
		this.y = y;
		this.I = I;
	}
}

// The Flash version has this special class because of the high overhead of creating/removing objects 
// Other languages solve this by using a "struct" for the Entry and "byref" uints for Dequeue
internal final class EntryPool
{
	private var _avail:Vector.<Entry>;
	private var _count:uint = 0;
	
	public final function EntryPool(capacity:uint)
	{
		this._avail = new Vector.<Entry>(capacity, true);
	}
	
	public final function Get(s:uint, index:uint, x:uint, y:uint, I:uint):Entry
	{
		if (this._count == 0)
			return new Entry(s, index, x, y, I);
		var e:Entry = this._avail[--this._count];
		e.score = s;
		e.index = index;
		e.x = x;
		e.y = y;
		e.I = I;
		return e;
	}
	
	public final function Return(e:Entry):void
	{
		if (this._count == this._avail.length)
		{
			this._avail.fixed = false;
			this._avail.length = this._count * 2;
			this._avail.fixed = true;
		}
		this._avail[this._count++] = e;
	}
}
